#!/usr/bin/perl
my $_script_build = 3;
my $_script_version = "0.1-${_script_build} (2014.04.06)";
#=======================================================================
# * snmp_sys_interrupts.pl
# * v0.1 (2015.04.06)
# * by red_neon
# * Monitoring of interrupts in Linux system. For network interfaces and other devices.
#=======================================================================
# INSTALLATION
#
#
# [REMOTE-SERVER SIDE]
#
#	[1] sys_interrupts script
#		Copy snmp_sys_interrupts.pl
#		from: sys_interrupts_v0.1/remote_server/
#		to: /etc/srv_scripts/ (create if not exists)
#
#	[2] snmpd
#	Put into snmpd.conf string:
#	$ nano /etc/snmp/snmpd.conf
#		pass_persist .1.3.6.1.4.1.2021.190.1000.1 /usr/bin/perl /etc/srv_scripts/snmp_sys_interrupts.pl
#
#	restart snmpd:
#		$ /etc/init.d/snmpd restart
#
#	checking:
#		$ snmpwalk -v 2c -c public 127.0.0.1 .1.3.6.1.4.1.2021.190.1000.1
#
#
# [CACTI-SERVER SIDE]
#
#	[3] cacti xml
#	Copy snmp_sys_interrupts.xml
#	from: sys_interrupts_v0.1/cacti_dir/resource/snmp_queries/
#	to: cacti/resource/snmp_queries/
#
#	[4] cacti templates
#	Import templates from sys_interrupts_v0.1/cacti_web_templates/ in Cacti:
#
#=======================================================================
# Sources:
# snmp.pl     http://www.docum.org/stef.coene/qos/gui/
# trafgraf.pl by Sébastien Cramatte <scramatte@zensoluciones.com> October 2006
#=======================================================================
use strict;
#=======================================================================
# Settings

my $_ip = "/sbin/ip";

# Descriptions max lenght
my $_desc_len = 18;

# Custom devices (separated with ",") example: "ehci_hcd:usb1, ath"
my $_my_devices = "";

my $_baseoid = '.1.3.6.1.4.1.2021.190.1000.1';

#=======================================================================
my $_debug = 0;
my $_cache_secs = 4;
my $_mibtime = 0;
my $_mib;
my %_oidnext_cache = ("upd_time" => 0, "oids" => {}, "oid_first" => 0); # oids - таблица: {OID => VALUE}
#=======================================================================
# Basic snmp functions
#=======================================================================
# Switch on autoflush
$| = 1;
my ($oid_in, $oid, $found);
while (my $cmd = <STDIN>) {
	chomp $cmd;
	if ($cmd eq "PING") {
		print "PONG\n";
	} elsif ($cmd eq "get") {
		$oid_in = <STDIN>;
		$oid = snmp_get_oid($oid_in); # .1.3.6.1.4.1.2021.190.1000.x -> x
		$_mib = create_mib("get");
		if (defined($_mib->{$oid})) {
			# ANSWER:
			# current oid
			# type of value
			# value
			print "$_baseoid.$oid\n";
			print $_mib->{$oid}[0]."\n";
			print $_mib->{$oid}[1]."\n";
		} else {
			print "NONE\n";
		}
	} elsif ($cmd eq "getnext") {
		$oid_in = <STDIN>;
		$oid = snmp_get_oid($oid_in);
		$_mib = create_mib("getnext");
		$found = 0;
		my $oidnext;
		# optimization
		if (exists $_oidnext_cache{"oids"}{$oid}) { # может быть ноль
			if ($_oidnext_cache{"oids"}{$oid}) { # если не ноль
				$oidnext = $_oidnext_cache{"oids"}{$oid};
				$found = 1;
			}
		} else {
			# если запрашиваемый ОИД не существует, то необходимо сделать walk всех наследников
			my @s = sort {snmp_oidcmp($a, $b)} keys %{$_mib};
			for my $i (0..(@s-1)) {
				if (snmp_oidcmp($oid, $s[$i]) == -1) {
					$oidnext = $s[$i];
					$found = 1;
					last;
				}
			}
		}
		if ($found) {
			# ANSWER:
			# next oid
			# type of value
			# value
			print "$_baseoid.".$oidnext."\n";
			print $_mib->{$oidnext}[0]."\n";
			print $_mib->{$oidnext}[1]."\n";
		} else {
			print "NONE\n";
		}
	} elsif ($cmd eq "getall") {
		$_mib = create_mib("get");
		my @s = sort {snmp_oidcmp($a, $b)} keys %{$_mib};
		for my $i (0..(@s-1)) {
			my $oid = $s[$i];
			print("$_baseoid.$oid = ".$_mib->{$oid}[0].": ".$_mib->{$oid}[1]."\n");
		}
	} else {
		# Unknown command
	}
}

exit 0;
#=======================================================================
sub snmp_get_oid {
#=======================================================================
	my ($oid) = @_;
	chomp($oid);
	my $base = $_baseoid;
	$base =~ s/\./\\./g;
	# Requested oid doesn't match base oid
	if ($oid !~ /^$base(\.|$)/) {
		return 0;
	}
	$oid =~ s/^$base\.?//;
	return $oid;
}
#=======================================================================
sub snmp_oidcmp {
#=======================================================================
	my ($x, $y) = @_;
	my @a = split(/\./, $x);
	my @b = split(/\./, $y);
	my $i = 0;
	while (1) {
		if ($i > $#a) {
			if ($i > $#b) {
				return 0;
			} else {
				return -1;
			}
		} elsif ($i > $#b) {
			return 1;
		}
		if ($a[$i] < $b[$i]) {
			return -1;
		} elsif ($a[$i] > $b[$i]) {
			return 1;
		}
		$i++;
	}
}
#=======================================================================
sub snmp_oidnext_cache {			# Функция кеширования (увеличивает скорость вывода snmpwalk)
#=======================================================================
	my ($snmp_res) = @_;
	my $maxn_mib = keys %{$snmp_res};
	my $maxn_cache = keys %{$_oidnext_cache{"oids"}};
	my $recache;
	# проверка актуальности кеша (елси он был сформирован, иначе - создать)
	# если размер оидов кеша соответствует размеру оидов массива - проверить на наличие всех пар индексов
	# если все текущие индексы (оиды) есть в кеше - значит незачем делать перекеширование
	if ($maxn_mib == $maxn_cache) {
		while ((my $oid) = each %{$snmp_res}) {
			if (!exists $_oidnext_cache{"oids"}{$oid}) { $recache = 1; last; }
		}
	# если размер не соответствует
	} else {
		$recache = 1;
	}
	# перекеширование
	if ($recache) {
		my $curtime = time();
		# сброс таблицы кеша
		%_oidnext_cache = ("upd_time" => $curtime, "oids" => {}, "oid_first" => 0);
		# сортируем массив в последовательности (1.0 - 1.1 - 2.0 - 2.1 и т.п.)
		my @s = sort {snmp_oidcmp($a, $b)} keys %{$snmp_res};
		# первый OID выносим отдельно
		if (exists $s[0]) { $_oidnext_cache{"oid_first"} = $s[0]; }
		for my $i (0..(@s-1)) {
			# создаём пары ({текущий} = следующий): $_oidnext_cache{"oids"}{'1.0'} = '1.1'
			# ноль будет иметь самый последний OID, у которого нет следующего!
			$_oidnext_cache{"oids"}{$s[$i]} = 0;
			if (exists $s[$i + 1]) {
				$_oidnext_cache{"oids"}{$s[$i]} = $s[$i + 1];
			}
		}
	}
}
#=======================================================================
sub create_mib {
#=======================================================================
	my ($arg) = @_;
	if (time - $_mibtime < $_cache_secs) { return $_mib; }

	my %ifaces = ();	# таблица интерфейсов
	my %snmp_res = ();	# таблица oids
	#################################
	my $on_exit; $on_exit = sub {
		$_mib = \%snmp_res;
		$_mibtime = time;
		if ($arg eq "getnext") {
			snmp_oidnext_cache($_mib);
		}
		return $_mib;
	};
	#################################
	# Ifaces table
	if (open(PROC, "$_ip link show 2>&1 |")) {
		my @output = <PROC>; close(PROC);
		foreach my $line (@output) {
			if ($line =~ m/^\s*\d+:\s+(\S+):.+\s+state\s+(\S+)\s+/) {
				my ($iface_name, $status) = ($1, $2);
				if ($iface_name eq 'lo' ) { next; }
				if (!exists $ifaces{$iface_name}) {
					if ($line =~ m/(NO\-CARRIER)/) { $status = $status." ($1)"; }
					$ifaces{$iface_name} = {'name' => $iface_name, 'irq' => "N/A", 'irq_tot' => "N/A", 'status' => $status, 'cpu_array' => [0]};
					if ($_debug) { print("IFACE: $iface_name $status \n"); }
				} else {
					$ifaces{$iface_name}{'status'} = $status;
				}
			}
		}
	}
	if (keys(%ifaces) == 0) {
		if ($_debug) { print("Interfaces not found via $_ip!\n"); }
	}
	# Custom ifaces
	my @userdevices = $_my_devices =~ m/([A-Za-z0-9_:;\%\$\@\{\}\[\]\(\)\+\-\/\.\\]+),?\s*/g;
	if ($_debug) { print("USER DEVICES: @userdevices \n"); }
	foreach my $dev (@userdevices) {
		if (!exists($ifaces{$dev})) {
			$ifaces{$dev} = {'name' => $dev, 'irq' => "N/A", 'irq_tot' => "N/A", 'status' => "N/A", 'cpu_array' => [0]};
		}
	}
	if (keys(%ifaces) == 0) {
		if ($_debug) { print("Interfaces not found!\n"); }
		return &$on_exit;
	}
	# Iface "ALL"
	$ifaces{"ALL"} = {'name' => "ALL", 'irq' => "ALL", 'irq_tot' => "N/A", 'status' => "OK", 'cpu_array' => [0]};
	#################################
	if (!open(FILE,'<','/proc/interrupts')) {
		if ($_debug) { print("Can't open interrupts file, error: $!\n"); }
		return &$on_exit;
	}
	my @output = <FILE>; close(FILE);
	my ($iline, $tot_cpu) = (0, 0);
	foreach my $line (@output) {
		chomp($line);
		$iline++;
		if ($line =~ m/\s+CPU(\d+)\s*$/) {
			$tot_cpu = $1 + 1;
			# заполняем массив с прерываниями по цпу нулями
			foreach my $iface_name (keys %ifaces) {
				for (my $i=0; $i < $tot_cpu; $i++) { $ifaces{$iface_name}{'cpu_array'}[$i] = 0; }
			}
		} else {
			if ($tot_cpu == 0) { next; }
			if ($line =~ m/^\s*(ERR:|MIS:)/) { next; }
			my ($iface_name, $iface_irq) = ('ALL', 'N/A');
			foreach my $iface (keys %ifaces) {
				if ($iface ne 'ALL' && ($line =~ m/^\s*(\d+):\s+.+\s+${iface}(\-[A-Za-z0-9_:;\%\$\@\{\}\[\]\(\)\+\-\/\.\\]+)?(,.*|\s+.*|)$/)) {
					($iface_name, $iface_irq) = ($iface, $1);
					last;
				}
			}
			if ($iface_name ne 'ALL' && $iface_irq ne 'N/A') {
				if ($ifaces{$iface_name}{'irq_tot'} eq 'N/A' ) { $ifaces{$iface_name}{'irq_tot'} = 0; }
				$ifaces{$iface_name}{'irq_tot'}++;
				if ($ifaces{$iface_name}{'irq'} eq 'N/A' ) { $ifaces{$iface_name}{'irq'} = $iface_irq; }
				else { $ifaces{$iface_name}{'irq'} = $ifaces{$iface_name}{'irq'}." ".$iface_irq; }
			}
			# Удаляем вначале номер irq:
			$line =~ s/^\s*\S+:\s+//;
			if ($_debug) { 
				$line =~ s/\s+/ /g;
				print("[$iline] line: [ $line ]\n");
			}
			for (my $i=0;$i<$tot_cpu;$i++) {
				if ($line =~ m/^(\d+)\s+/) {
					my $val = $1;
					my $oldval = $ifaces{$iface_name}{'cpu_array'}[$i] ;
					if ($iface_name ne 'ALL') {
						$ifaces{$iface_name}{'cpu_array'}[$i] = $ifaces{$iface_name}{'cpu_array'}[$i] + $val;
					}
					$ifaces{'ALL'}{'cpu_array'}[$i] = $ifaces{'ALL'}{'cpu_array'}[$i] + $val;
					if ($_debug) { print("[$iline] [$iface_name] [CPU$i] current+old=... [$val]+[$oldval]=[$ifaces{$iface_name}{'cpu_array'}[$i]]\n"); }
					# удаляем первое значение
					$line =~ s/^\s*\S+\s+//;
					if ($_debug) {
						if ($i == $tot_cpu-1) { print("[$iline] \t\t[END] newline: [ $line ]\n"); }
						else { print("[$iline] \t\tnewline: [ $line ]\n"); }
					}
				} else {
					if ($_debug) { print("[$iline] Error: value for current cpu #$i is not found!\n"); }
				}
			}
		}
	}
	#################################
	%snmp_res = (
		# Общая информация (структура индекс-подобная, статическая, с индексом 1):
		"1.0" => [ "Integer", 1 ],						# Index (всегда единица)
		"1.0.1" => [ "Integer", 1 ],					# Index
		"1.1.1" => [ "String", "SYS INTERRUPTS v".$_script_version ],# Script Description
		"1.2.1" => [ "Gauge32", $_script_build ],		# Current Build (increment)

		# Таблица интерфейсов
		"2.0" => [ "Integer", "".keys(%ifaces)."" ],	# total ifaces

		#"3.x.x" => [ "Counter32", x ],					# interrupts count on cpu x on iface x 
	);
	
	my $iface_index = 0;
	foreach my $iface_name (sort {$ifaces{$a}{"name"} cmp $ifaces{$b}{"name"}} keys %ifaces) {
		$iface_index++;
		$snmp_res{"2.1.$iface_index"} = [ "Integer", $iface_index ];						# index
		$snmp_res{"2.2.$iface_index"} = [ "String", $iface_name ];							# iface name
		
		my $iface_desc = $ifaces{$iface_name}{'irq_tot'}.' IRQ';
		if ($ifaces{$iface_name}{'irq_tot'} eq 'N/A' || $ifaces{$iface_name}{'irq_tot'} gt 1) { $iface_desc = $iface_desc.'s'; }
		$iface_desc = $iface_desc.': '.$ifaces{$iface_name}{'irq'};
		if ($iface_name eq 'ALL') { $iface_desc = 'ALL IRQs'; }
		elsif ($ifaces{$iface_name}{'irq'} eq 'N/A') { $iface_desc = 'IRQs not found'; }
		if (length($iface_desc) > $_desc_len) {
			if ($iface_desc =~ m/^(.{$_desc_len})/) { $iface_desc = $1.".." };
		}
		if ($ifaces{$iface_name}{'status'} eq 'N/A' && $ifaces{$iface_name}{'irq_tot'} > 0) {
			$ifaces{$iface_name}{'status'} = 'OK';
		}

		$snmp_res{"2.3.$iface_index"} = [ "String", $ifaces{$iface_name}{'status'} ];		# iface status
		$snmp_res{"2.4.$iface_index"} = [ "String", $ifaces{$iface_name}{'irq'} ];			# all irqs in format: 100 101 102
		$snmp_res{"2.5.$iface_index"} = [ "String", $ifaces{$iface_name}{'irq_tot'} ];		# total irqs number
		$snmp_res{"2.6.$iface_index"} = [ "String", $iface_desc ];								# desc, total irq + numbers
		$snmp_res{"2.7.$iface_index"} = [ "String", "[$tot_cpu] CPUs" ];					# type of graph
		if ($_debug) { print("iface name: [ $iface_name ] \$snmp_res{\"1.3.$iface_index\"} = [ \"String\", $ifaces{$iface_name}{'irq'} ]\n"); }

		if ($iface_name ne 'ALL' && $ifaces{$iface_name}{'irq_tot'} =~ m/^(0|N\/A)$/) { next; }

		for (my $i=0; $i<$tot_cpu; $i++) {
			$snmp_res{"3.".$i.".".$iface_index} = [ "Counter32", adjust_to_32bit($ifaces{$iface_name}{'cpu_array'}[$i]) ];
			if ($_debug) { print("iface name: [ $iface_name ] [CPU$i] \$snmp_res{3.\"$i.$iface_index\"} = [ \"Counter32\", ".adjust_to_32bit($ifaces{$iface_name}{'cpu_array'}[$i])." ]\n"); }
		}
	}
	#################################
	return &$on_exit;
	#################################
} 
#=======================================================================
sub adjust_to_32bit {
#=======================================================================
	my ($val) = @_;
	if ($val > 4294967295) { $val = $val % 4294967295; }
	return $val;
}
#=======================================================================
